G_cgs = 6.674e-8  # gravitational constant in cgs
c_cgs = 2.998e10  # speed of light in cgs
M_sun_cgs = 1.989e33  # solar mass in cgs
Mpc2cm = 3.086e24  # from Megaparsec to centimeters
Mpc2km = 3.086e19  # from Megaparsec to centimeters



import numpy as np
import healpy as hp
import os
from scipy import stats, interpolate
from .cosmology import d_c, z_from_d_c
from .constants import c_cgs, Mpc2cm, Mpc2km


def load_Pk_gadget(simDir, nsnap, nbins=None, logbins=True, getz=False, fout=None):
    """Loads the 3D matter power spectrum from a Gadget-4 simulation.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder.
    nsnap : int
        Number of the shapshot.
    nbins : int
        Number of k bins.
    logbins : bool
        Whether the k bins are log spaced or not.
    getz : bool
        Returns in addition the snapshot redshift.
    fout : str
        Path to the folder in which to save the data.

    Returns
    -------
    kvals : array
        Array containing the values of the wavenumber k on which the
        bins are centered.
    Pk : array
        Array containing P(k).
    z: float
        The z of the power spectrum (returned only if getz is true).
    """
    # Look for the file
    found = False
    for file in os.listdir(simDir + "/powerspecs"):
        fname = os.fsdecode(file)
        if fname.endswith(f"_{nsnap:03d}.txt"):
            found = True
            break
    if not (found):
        print("File not found!")
        return
    # Read the file
    fpath = simDir + "/powerspecs/" + fname
    f = open(fpath)
    lines = f.readlines()
    nbins_file = int(lines[1])
    kvals, Pk = [], []
    z = 1 / float(lines[0]) - 1
    for i in range(5, nbins_file + 5):
        idx = [j for j, letter in enumerate(lines[i]) if letter == " "]
        k = float(lines[i][: idx[0]])
        kvals.append(k)
        delta = float(lines[i][idx[0] : idx[1]])
        Pk.append(2 * (np.pi) ** 2 * delta / k**3)
    # Create bins for k
    if nbins:
        if logbins:
            kbins = np.logspace(np.log10(kvals[0]), np.log10(kvals[-1]), nbins + 1)
        else:
            kbins = np.linspace(kvals[0], kvals[-1], nbins + 1)
        Pk, _, _ = stats.binned_statistic(kvals, Pk, statistic="mean", bins=kbins)
        kvals = 0.5 * (kbins[1:] + kbins[:-1])
    # Eventually save Pk in numpy format
    if fout:
        print(f"Saved Pk to {fout}_Pk")
        np.save(fout + "_Pk", np.stack((kvals, Pk), axis=-1))
    if getz:
        return kvals, Pk, z
    else:
        return kvals, Pk


def Pk_2D(m, lside, nbins=20, logbins=True, zero_padding=False, fout=None):
    """Computes the P(k) of a 2D density field. This field is given as input in the
    square matrix m, which must be a np array.

    Parameters
    ----------
    m : 2D array
        Square map on which to compute the power spectrum.
    lside : float
        Physical lenght of the side of the map.
    nbins : int
        Number of k bins.
    logbins : bool
        Whether the l bins are log spaced or not.
    zero_padding : bool:
        Whether to apply zero padding or not
    fout : str
        Path to the folder in which to save the data.

    Returns
    -------
    kvals : array
        Array containing the values of the wavenumber k on which the
        bins are centered.
    Abins : array
        Array containing P(k).
    """
    n = m.shape[0] if not (zero_padding) else 2 * m.shape[0]
    lside = lside if not (zero_padding) else 2 * lside
    # Compute fundamental frequency
    kF = 2.0 * np.pi / lside
    # Compute the Fourier transform of m
    if not (zero_padding):
        fourier_m = np.fft.fftn(m)
    else:
        m_pad = np.zeros([n, n], dtype=float)
        m_pad[: m.shape[0], : m.shape[0]] += m
        fourier_m = np.fft.fftn(m_pad)
    # Compute the amplitude of the Fourier components
    fourier_amp = np.abs(fourier_m) ** 2
    # Construct the wave vector array
    kfreq = np.fft.fftfreq(n) * n
    kfreq2D = np.meshgrid(kfreq, kfreq)
    # We are interested in the norm of the wave vectors
    knrm = np.sqrt(kfreq2D[0] ** 2 + kfreq2D[1] ** 2)
    # We can now flatten the 2D arrays
    knrm, fourier_amp = knrm.flatten(), fourier_amp.flatten()
    # Create bins for k
    k_start = 0.5 if not (zero_padding) else 1.0
    if logbins:
        kbins = np.logspace(np.log10(k_start), np.log10(n // 2 + 1), nbins + 1)
    else:
        kbins = np.linspace(k_start, n // 2 + 1, nbins + 1)
    kvals = 0.5 * (kbins[1:] + kbins[:-1])
    Abins, _, _ = stats.binned_statistic(
        knrm, fourier_amp, statistic="mean", bins=kbins
    )
    if zero_padding:
        Abins *= 4
    # Give units
    for i in range(len(kvals)):
        kvals[i] = (kvals[i]) * kF
        Abins[i] = (Abins[i]) * (lside / n**2) ** 2
    # Eventually save Pk in numpy format
    if fout:
        print(f"Saved Pk to {fout}_Pk")
        np.save(fout + "_Pk", np.stack((kvals, Abins), axis=-1))
    return np.array([kvals, Abins])


def C_ell(m, lmax=None, nest=False, nbins=None, logbins=True, fout=None):
    """Computes the angular power spectrum (known as C(l)) of the map m through the
    healpy anafast routine.

    Parameters
    ----------
    m : array
        HEALPix map on which to compute the power spectrum.
    lmax : float
        Maximum l of the power spectrum (default: 3*nside-1).
    nest : bool
        Whether the map is nest ordered or not.
    nbins : int
        Number of l bins.
    logbins : bool
        Whether the l bins are log spaced or not.
    fout : str
        Path to the folder in which to save the data.

    Returns
    -------
    l : array
        Array containing the values of l.
    cl : array
        Array containing C(l).
    """
    rho = m
    # Map must be ring ordered
    if nest:
        rho = hp.reorder(rho, n2r=True)
    # Compute the power spectrum
    if lmax == None:
        Cl = hp.anafast(rho, use_weights=True)
    else:
        Cl = hp.anafast(rho, use_weights=True, lmax=lmax)
    ll = np.arange(len(Cl))
    # Eventually rebin
    if nbins:
        if logbins:
            lbins = np.logspace(np.log10(1), np.log10(ll[-1]), nbins + 1)
        else:
            lbins = np.linspace(ll[0], ll[-1], nbins + 1)
        Cl, _, _ = stats.binned_statistic(ll, Cl, statistic="mean", bins=lbins)
        ll = 0.5 * (lbins[1:] + lbins[:-1])
    # Eventually save C(l) in numpy format
    if fout:
        print(f"Saved Cl to {fout}_Cl")
        np.save(fout + "_Cl", np.stack((ll, Cl), axis=-1))
    return ll, Cl


import numpy as np
from scipy import ndimage


def smooth_region(m, sigma, lside, fout=None):
    """Convolves a square map with a gaussian kernel. Note that sigma and lside
    must be given in the same units.

    Parameters
    ----------
    m : 2D array
        Image on which to smooth.
    sigma : float
        Standard deviation of the kernel.
    lside : float
        Physical lenght of the side of the image.
    fout : str
        Path of the folder which to save the output.

    Returns
    -------
    2D array
        The smoothed map.
    """
    sigma_0 = sigma / (lside / m.shape[0])
    m_smooth = ndimage.filters.gaussian_filter(m, sigma_0, mode="constant")
    # Eventually save smoothed region in numpy format
    if fout:
        print(f"Saved smoothed region to {fout}_smoothed")
        np.save(fout + "_smoothed", m_smooth)
    return m_smooth


def add_gaussian_noise(m, sigma_e, n_gal, A_pix, fout=None):
    """Adds gaussian shape noise to a square map, as modeled in eq. (7) of
    Lee+ 22 https://arxiv.org/abs/2201.08320.

    Parameters
    ----------
    m : 2D-array or HEALPix map
        Image on which to add the gaussian noise.
    sigma_e : float
        Intrinsic ellipticity of galaxies.
    n_gal : float
        Surface number density of lensed galaxies in arcmin^-2.
    sigma_e : float
        Intrinsical ellipticity of galaxies.
    A_pix : float
        Solid angle of a pixel in arcmin^2.
    fout : str
        Path of the folder which to save the output.

    Returns
    -------
    array-like
        Image with gaussian noise added.
    """
    sigma = sigma_e / np.sqrt(2 * n_gal * A_pix)
    noise = np.random.normal(0, sigma, m.shape)
    m_noise = m + noise
    # Eventually save region with noise in numpy format
    if fout:
        print(f"Saved region with noise to {fout}_noise")
        np.save(fout + "_noise", m_noise)
    return m_noise


def fibonacci_sphere(n=1000, lonlat=True):
    """Distribute n points on a sphere with the Fibonacci algorithm, returns
    the positions of such points.
    More on the algorithm at:
    https://extremelearning.com.au/evenly-distributing-points-on-a-sphere/

    Parameters
    ----------
    n : int
        Number of points to place on the sphere.
    lonlat : bool
        wheter the output has to be as longitude and latitude in degrees or
        colatitude and longitude in radians.

    Returns
    -------
    2D array
        Array with shape n x 2, each row representing a point.
    """
    goldenRatio = (1 + 5**0.5) / 2
    i = np.arange(0, n)
    phi = (2 * np.pi * i / goldenRatio) % (2 * np.pi)
    theta = np.arccos(1 - 2 * (i + 0.5) / n)
    if lonlat:
        coords = np.stack((np.degrees(phi), np.degrees(theta - np.pi / 2)), axis=-1)
    else:
        coords = np.stack((theta, phi), axis=-1)
    return coords


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
        # Optional: return something else or raise an error if called again
    wrapper.has_run = False
    return wrapper

@run_once
def print_logo():
    logo = """
    ▐▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▌
    ▐                                                               ▌
    ▐   ██████████                       ███                        ▌
    ▐  ░░███░░░░███                     ░░░                         ▌
    ▐   ░███   ░░███  ██████  ████████  ████   ██████   ████████    ▌
    ▐   ░███    ░███ ███░░███░░███░░███░░███  ░░░░░███ ░░███░░███   ▌
    ▐   ░███    ░███░███ ░███ ░███ ░░░  ░███   ███████  ░███ ░███   ▌
    ▐   ░███    ███ ░███ ░███ ░███      ░███  ███░░███  ░███ ░███   ▌
    ▐   ██████████  ░░██████  █████     █████░░████████ ████ █████  ▌
    ▐  ░░░░░░░░░░    ░░░░░░  ░░░░░     ░░░░░  ░░░░░░░░ ░░░░ ░░░░░   ▌
    ▐                                                               ▌
    ▐▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▌
    """
    print(logo)


"""Coloured terminal logging utilities for Dorian.

Provides ``info``, ``success``, ``warning``, ``error``, ``debug``, and
``banner`` helpers. The module-level variable ``_prefix`` (default ``""``) is
prepended to every message and can be set per-process to tag output from
parallel workers (e.g., ``_prefix = "[z_s=1.00] "``).
"""
import sys


class Colors:
    """ANSI color codes for terminal output, with TTY detection."""

    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    DIM = "\033[2m"
    BOLD = "\033[1m"
    RESET = "\033[0m"

    _enabled = True

    @classmethod
    def _check_tty(cls):
        if not sys.stdout.isatty():
            cls.disable()

    @classmethod
    def disable(cls):
        cls._enabled = False
        cls.BLUE = ""
        cls.GREEN = ""
        cls.YELLOW = ""
        cls.RED = ""
        cls.DIM = ""
        cls.BOLD = ""
        cls.RESET = ""


Colors._check_tty()

# Module-level prefix, can be set per-process for parallel workers
_prefix: str = ""


def info(message):
    """Print an info message with blue [INFO] prefix."""
    print(f"{Colors.BLUE}[INFO]{Colors.RESET} {_prefix}{message}")


def success(message):
    """Print a success message with green checkmark."""
    print(f"{Colors.GREEN}\u2714{Colors.RESET} {_prefix}{message}")


def warning(message):
    """Print a warning message with yellow [WARNING] prefix."""
    print(f"{Colors.YELLOW}[WARNING]{Colors.RESET} {_prefix}{message}")


def error(message):
    """Print an error message with red [ERROR] prefix to stderr."""
    print(f"{Colors.RED}[ERROR]{Colors.RESET} {_prefix}{message}", file=sys.stderr)


def debug(message):
    """Print a debug message with dim [DEBUG] prefix and separator lines."""
    sep = Colors.DIM + "-" * 40 + Colors.RESET
    print(f"{sep}\n{Colors.DIM}[DEBUG]{Colors.RESET} {_prefix}{message}\n{sep}")


def banner(message):
    """Print a separator banner."""
    line = "=" * 60
    print(f"{Colors.BOLD}{line}\n  {_prefix}{message}\n{line}{Colors.RESET}")



from .constants import M_sun_cgs, Mpc2cm, c_cgs, G_cgs
from .cosmology import d_c
from .parallel_transport import get_rotation_angle_array, rotate_tensor_array
from .misc import print_logo
from .logging import info, success, warning
import numpy as np
import healpy as hp
from ducc0.sht import synthesis_general
import time


def raytrace(
    shells,
    z_s,
    omega_m,
    omega_l,
    nside,
    shell_redshifts,
    shell_distances,
    interp="bilinear",
    lmax=0,
    parallel_transport=True,
    nthreads=1,
):
    """
    Perform full-sky gravitational lensing ray-tracing through multiple lens planes.

    This is the core ray-tracing algorithm in Dorian. It propagates light rays
    backwards from the observer through a series of mass shells, computing
    deflections and the Jacobian distortion matrix at each step.

    Parameters
    ----------
    shells : list of np.ndarray
        List of HEALPix mass maps for each shell. Each map must contain mass
        per pixel in units of 10^10 M_sun/h. Use ``prepare_density_shells()``
        to convert from other formats. Maps should be in RING ordering.
    z_s : float
        Source redshift. Only shells with ``z < z_s`` will contribute.
    omega_m : float
        Matter density parameter (Omega_m).
    omega_l : float
        Dark energy density parameter (Omega_Lambda).
    nside : int
        HEALPix NSIDE parameter defining the angular resolution.
        Number of pixels = ``12 * nside**2``.
    shell_redshifts : list of float
        Redshift of each shell, same length as ``shells``.
    shell_distances : list of float
        Comoving distance to each shell in Mpc/h.
    interp : {'bilinear', 'ngp', 'nufft'}, optional
        Interpolation method for sampling fields at ray positions:

        - ``'bilinear'``: Bilinear interpolation (default, recommended)
        - ``'ngp'``: Nearest grid point (fastest, lowest accuracy)
        - ``'nufft'``: Non-uniform FFT via ducc0 (highest accuracy)

    lmax : int, optional
        Maximum multipole ell for spherical harmonic transforms.
        Default: ``3 * nside`` when set to 0.
    parallel_transport : bool, optional
        Whether to parallel transport the distortion matrix along geodesics
        when rays are deflected. Recommended True for accurate results.
        Default: True.
    nthreads : int, optional
        Number of OpenMP threads for ``'nufft'`` interpolation. Default: 1.

    Returns
    -------
    kappa_born : np.ndarray
        Born approximation convergence map. Shape: ``(npix,)``.
    A_final : np.ndarray
        Final Jacobian distortion matrix. Shape: ``(2, 2, npix)``.
        Components: A[0,0]=A_11, A[0,1]=A_12, A[1,0]=A_21, A[1,1]=A_22.
    beta_final : np.ndarray
        Final angular positions (theta, phi) of rays. Shape: ``(2, npix)``.
        theta in [0, pi], phi in [0, 2*pi].
    theta : np.ndarray
        Initial ray positions (HEALPix pixel centers). Shape: ``(2, npix)``.

    Notes
    -----
    The algorithm iterates through shells from observer outward:

    1. Compute surface density Sigma and convergence kappa for each shell
    2. Transform kappa to spherical harmonics to get deflection angles alpha
    3. Interpolate alpha at current ray positions (post-Born correction)
    4. Update ray positions using the multi-plane lens equation
    5. Update distortion matrix A with the tidal tensor U
    6. Optionally parallel transport A to the new ray direction
    7. Accumulate Born approximation convergence

    The distortion matrix A relates source to image coordinates:
    ``d(beta)/d(theta) = A``, initialized to identity.

    Lensing observables from A:
        - Convergence: ``kappa = 1 - 0.5 * (A[0,0] + A[1,1])``
        - Shear: ``gamma1 = 0.5 * (A[0,0] - A[1,1])``, ``gamma2 = A[0,1]``

    Examples
    --------
    Low-level usage (prefer ``raytrace_from_density`` for convenience):

    >>> from dorian.raytracing import raytrace
    >>> from dorian.lensing import prepare_density_shells
    >>> # Prepare shells first
    >>> shells, distances, redshifts = prepare_density_shells(...)
    >>> # Run ray-tracing
    >>> kappa_born, A, beta, theta = raytrace(
    ...     shells=shells,
    ...     z_s=1.0,
    ...     omega_m=0.3,
    ...     omega_l=0.7,
    ...     nside=512,
    ...     shell_redshifts=redshifts,
    ...     shell_distances=distances,
    ... )
    >>> # Compute ray-traced convergence
    >>> kappa_rt = 1.0 - 0.5 * (A[0, 0] + A[1, 1])

    See Also
    --------
    dorian.lensing.raytrace_from_density : High-level interface (recommended).
    dorian.lensing.prepare_density_shells : Convert density maps to mass format.
    """
    t_begin = time.time()
    kappa_fac = (1e10 * M_sun_cgs) * (1 / Mpc2cm) * 4 * np.pi * G_cgs / (c_cgs**2)

    contributing_shells = []
    for i, (shell, z_k, d_k) in enumerate(zip(shells, shell_redshifts, shell_distances)):
        if z_k < z_s:
            contributing_shells.append({
                'shell_data': shell,
                'redshift': z_k,
                'distance': d_k,
                'index': i
            })

    n_contrib = len(contributing_shells)
    npix = hp.nside2npix(nside)
    info(f"Ray-tracing: nside={nside}, z_source={z_s}, interp='{interp}', "
         f"{n_contrib}/{len(shells)} shells contributing")

    if n_contrib == 0:
        warning(f"No shells contribute (all z >= z_source={z_s})")
        theta = np.array(hp.pix2ang(nside, np.arange(npix)))
        nrays = theta.shape[1]

        # kappa_born is correctly zero (no matter, no convergence)
        kappa_born = np.zeros(nrays)

        # beta (final positions) equals initial theta
        beta_final = theta.copy()

        # A (distortion matrix) must be the IDENTITY matrix, not zeros
        # A_final shape is (2, 2, npix)
        A_final = np.zeros((2, 2, nrays))
        A_final[0, 0, :] = 1.0
        A_final[1, 1, :] = 1.0

        return kappa_born, A_final, beta_final, theta

    d_s = d_c(z=z_s, Omega_M=omega_m, Omega_L=omega_l)

    theta = np.array(hp.pix2ang(nside, np.arange(npix)))
    nrays = theta.shape[1]
    beta = np.zeros([2, 2, nrays])
    A = np.zeros([2, 2, 2, nrays])
    kappa_born = np.zeros([nrays])

    beta[0] = theta
    beta[1] = theta
    for i in range(2):
        for j in range(2):
            A[0][i][j] = 1 if i == j else 0
            A[1][i][j] = 1 if i == j else 0
    sh_start = 0

    if lmax==0:
        lmax = 3 * nside
    ell = np.arange(0, lmax + 1)

    for k in range(sh_start, len(contributing_shells)):
        t0 = time.time()
        shell_info = contributing_shells[k]
        z_k = shell_info['redshift']
        d_k = shell_info['distance']
        shell_data = shell_info['shell_data']
        info(f"Shell {k+1}/{n_contrib}: z={z_k:.4f}, d={d_k:.1f} Mpc/h")

        Sigma = shell_data / (4 * np.pi / npix)
        Sigma_mean = np.mean(Sigma)

        kappa = kappa_fac * (1 + z_k) * (1 / d_k) * (Sigma - Sigma_mean)

        kappa_lm = hp.map2alm(kappa, pol=False, lmax=lmax)
        alpha_factor = (np.sqrt((ell * (ell + 1)))) # Safe-divide: guarded below at ell=0
        alpha_factor = np.where(alpha_factor == 0, 1.0, alpha_factor) # Avoid division by zero at ell=0
        alpha_lm = hp.almxfl(kappa_lm, -2 / alpha_factor)
        fl_factor = (ell * (ell + 1.0))
        fl_factor = np.where(fl_factor == 0, 1.0, fl_factor) # Avoid division by zero at ell=0
        f_l = -np.sqrt(np.maximum(0.0, (ell + 2.0) * (ell - 1.0) / fl_factor)) # Must safe divide; clamp ell=0 (monopole tidal = 0)
        g_lm_E = hp.almxfl(kappa_lm, f_l)

        if interp in ["ngp", "bilinear"]:
            alpha = hp.alm2map_spin(
                [alpha_lm, np.zeros_like(alpha_lm)], nside=nside, spin=1, lmax=lmax
            )
            alpha = get_val(alpha, beta[1][0], beta[1][1], interp=interp)

            g1, g2 = hp.alm2map_spin(
                [g_lm_E, np.zeros_like(g_lm_E)], nside=nside, spin=2, lmax=lmax
            )
            U = np.zeros([2, 2, nrays])
            U[0][0] = kappa + g1
            U[1][0] = g2
            U[0][1] = U[1][0]
            U[1][1] = kappa - g1
            U[0, 0], U[0, 1], U[1, 1] = get_val(
                [U[0, 0], U[0, 1], U[1, 1]], beta[1][0], beta[1][1], interp=interp
            )
            U[1, 0] = U[0, 1]

        elif interp == "nufft":
            alpha = get_val_nufft(
                alpha_lm, beta[1][0], beta[1][1], spin=1, lmax=lmax, nthreads=nthreads
            )
            g1, g2 = get_val_nufft(
                g_lm_E, beta[1][0], beta[1][1], spin=2, lmax=lmax, nthreads=nthreads
            )
            kappa_nufft = get_val_nufft(
                kappa_lm, beta[1][0], beta[1][1], spin=0, lmax=lmax, nthreads=nthreads
            )[0]

            U = np.zeros([2, 2, nrays])
            U[0][0] = kappa_nufft + g1
            U[1][0] = g2
            U[0][1] = U[1][0]
            U[1][1] = kappa_nufft - g1

        d_km1 = 0 if k==0 else contributing_shells[k-1]['distance']
        d_kp1 = d_s if k == len(contributing_shells) - 1 else contributing_shells[k+1]['distance']
        fac1 = d_k/d_kp1 * (d_kp1-d_km1)/(d_k-d_km1)
        fac2 = (d_kp1-d_k)/d_kp1

        for i in range(2):
            beta[0][i] = (1 - fac1) * beta[0][i] + fac1 * beta[1][i] - fac2 * alpha[i]

        beta[[0, 1], ...] = beta[[1, 0], ...]

        check_theta_poles(beta[1])
        beta[1][1] %= 2 * np.pi

        for i in range(2):
            for j in range(2):
                A[0][i][j] = (
                    (1 - fac1) * A[0][i][j]
                    + fac1 * A[1][i][j]
                    - fac2 * (U[i][0] * A[1][0][j] + U[i][1] * A[1][1][j])
                )

        A[[0, 1], ...] = A[[1, 0], ...]

        if parallel_transport:

            cospsi, sinpsi = get_rotation_angle_array(
                beta[0][0][:], beta[0][1][:], beta[1][0][:], beta[1][1][:]
            )
            A[0, :, :, :] = rotate_tensor_array(A[0, :, :, :], cospsi, sinpsi)
            A[1, :, :, :] = rotate_tensor_array(A[1, :, :, :], cospsi, sinpsi)

        kappa_born += ((d_s - d_k) / d_s) * kappa

        t1 = time.time()
        info(f"  Shell {k+1}/{n_contrib} done in {t1 - t0:.2f}s")

    t_end = time.time()
    success(f"Ray-tracing complete: {n_contrib} shells in {t_end - t_begin:.2f}s")

    return kappa_born, A[1], beta[1], theta


def get_val(m_list, theta, phi, interp):
    """
    Interpolate HEALPix maps at arbitrary angular positions.

    Parameters
    ----------
    m_list : list of np.ndarray
        List of HEALPix maps to interpolate. All maps must have the same NSIDE.
    theta : np.ndarray
        Co-latitude coordinates in radians, range [0, pi]. Shape: ``(npoints,)``.
    phi : np.ndarray
        Longitude coordinates in radians, range [0, 2*pi]. Shape: ``(npoints,)``.
    interp : {'ngp', 'bilinear'}
        Interpolation method:

        - ``'ngp'``: Nearest grid point (fastest)
        - ``'bilinear'``: Bilinear interpolation using 4 nearest pixels

    Returns
    -------
    list of np.ndarray
        Interpolated values for each input map. Each array has shape ``(npoints,)``.
    """
    nside = hp.npix2nside(len(m_list[0]))
    if interp == "ngp":
        idx = hp.ang2pix(nside, theta, phi)
        return [m[idx] for m in m_list]
    if interp == "bilinear":
        p, w = hp.get_interp_weights(nside, theta, phi)
        return [np.sum(m[p] * w, 0) for m in m_list]


def get_val_nufft(alm, theta, phi, spin, lmax, nthreads):
    """
    Evaluate spherical harmonic expansion at arbitrary positions using NUFFT.

    Uses ducc0's ``synthesis_general`` for high-accuracy non-uniform evaluation
    of spin-weighted spherical harmonic transforms.

    Parameters
    ----------
    alm : np.ndarray
        Spherical harmonic coefficients in healpy's indexing convention.
    theta : np.ndarray
        Co-latitude coordinates in radians, range [0, pi]. Shape: ``(npoints,)``.
    phi : np.ndarray
        Longitude coordinates in radians, range [0, 2*pi]. Shape: ``(npoints,)``.
    spin : int
        Spin weight of the field (0 for scalars, 1 for vectors, 2 for tensors).
    lmax : int
        Maximum multipole ell in the expansion.
    nthreads : int
        Number of OpenMP threads for parallel computation.

    Returns
    -------
    np.ndarray
        Evaluated field values at the specified positions.
        For spin > 0, returns array with shape ``(2, npoints)`` for the two
        spin components.
    """
    if spin == 0:
        alm2 = alm.reshape((1, -1))
    elif spin > 0:
        alm2 = np.zeros((2, alm.shape[0]), dtype=alm.dtype)
        alm2[0] = alm
    return synthesis_general(
        alm=alm2, spin=spin, lmax=lmax, loc=np.vstack([theta, phi]).T, nthreads=nthreads
    )


def check_theta_poles(coords):
    """
    Ensure theta coordinates remain within valid range [0, pi] near poles.

    After ray deflection, some rays near the poles may have theta values
    outside the valid range. This function corrects them by reflecting
    across the poles and adjusting phi accordingly.

    Only checks rays near the poles (0.5% at each end) assuming HEALPix
    RING ordering where polar rays are at the start and end of the array.

    Parameters
    ----------
    coords : np.ndarray
        Ray coordinates with shape ``(2, nrays)``. First row is theta
        (co-latitude), second row is phi (longitude). Modified in-place.

    Notes
    -----
    Correction rules:

    - If theta < 0: reflect to -theta, shift phi by pi
    - If theta > pi: reflect to 2*pi - theta, shift phi by pi
    """
    n_check = int(0.005 * len(coords[:]))
    coords_pole_north = coords[:, :n_check]
    coords_pole_south = coords[:, -n_check:]
    for coords_pole in [coords_pole_north, coords_pole_south]:
        idx = np.where(coords_pole[0] < 0)
        coords_pole[0, idx] = -coords_pole[0, idx]
        idx = np.where(coords_pole[0] > np.pi)
        coords_pole[0, idx] = 2 * np.pi - coords_pole[0, idx]
        coords_pole[1, idx] += np.pi
        coords_pole[1, idx] %= 2 * np.pi
    coords[:, :n_check] = coords_pole_north
    coords[:, -n_check:] = coords_pole_south



import numpy as np
import healpy as hp


def parallel_transport_sphere(tv, v_theta, v_phi, w_theta, w_phi, tv_is_vector=False):
    """Parallel transports the tangent tensor (or vector) tv on the surface of a sphere
    along the geodesic connecting the position vectors v and w.
    (i.e. parallel transports tv from v to w)

    Parameters
    ----------
    tv : 2D-array or 2x2 matrix
        Tangent vector/tensor at position v to be parallel transported.
    v_theta : float
        Colatitude of initial position.
    v_phi : float
        Longitude of initial position.
    w_theta : float
        Colatitude of final position.
    w_phi : float
        Longitude of final position.
    tv_is_vector : bool
        Whether tv is a vector (i.e. 1st degree tensor).

    Returns
    -------
    float
        Parallel transported tangent tensor (vector) at position w.
    """
    # convert angles to 3D position vector
    v = hp.ang2vec(v_theta, v_phi)
    w = hp.ang2vec(w_theta, w_phi)

    # compute axis
    axis = np.cross(v, w)
    cosangle = np.dot(v, w)
    sinangle = np.linalg.norm(axis)
    axis = axis / sinangle if (sinangle != 0) else np.array([1.0, 0.0, 0.0])

    # define a vector which lays on the equator and is perpendicular to v
    p = np.array([-v[1], v[0], 0.0])
    # the following rotates p around axis by an angle defined by singangle and cosangle (rodrigues formula)
    rephi_v = (
        p * cosangle
        + np.cross(axis, p) * sinangle
        + axis * np.dot(axis, p) * (1 - cosangle)
    )

    # define ephi_w and etheta_w
    ephi_w = np.array([-w[1], w[0], 0.0])
    etheta_w = np.array([w[2] * w[0], w[2] * w[1], -1.0 * (w[0] ** 2 + w[1] ** 2)])

    # compute the final rotation angle
    norm = np.sqrt((1.0 - w[2]) * (1.0 + w[2]) * (1.0 - v[2]) * (1.0 + v[2]))
    if norm == 0.0:
        sinpsi = 0.0
        cospsi = 1.0
    else:
        sinpsi = (
            rephi_v[0] * etheta_w[0]
            + rephi_v[1] * etheta_w[1]
            + rephi_v[2] * etheta_w[2]
        ) / norm
        cospsi = (
            rephi_v[0] * ephi_w[0] + rephi_v[1] * ephi_w[1] + rephi_v[2] * ephi_w[2]
        ) / norm

    if tv_is_vector:
        # compute the parallel transported vector
        tw = np.zeros_like(tv)
        tw[0] = tv[0] * cospsi + tv[1] * sinpsi
        tw[1] = -1.0 * tv[0] * sinpsi + tv[1] * cospsi
    else:
        # compute the parallel transported tensor
        R = np.array([[cospsi, sinpsi], [-sinpsi, cospsi]])
        tw = R.dot(tv).dot(R.T)
    return tw


def parallel_transport_sphere_array(
    tv, v_theta, v_phi, w_theta, w_phi, tv_is_vector=False
):
    """Same as parallel_transport_sphere, but parallel transports an array of
    tensors (vectors) from the positions v to the positions w.
    In this case tv must be an array od N elements, and consequently v and w must be
    an array of N vectors.

    Parameters
    ----------
    tv : 1D array of 2D or 1D arrays (float)
        Tangent tensors (vectors) at position v to be parallel transported.
    v_theta : 1D array (float)
        Colatitudes of initial position.
    v_phi : 1D array (float)
        Longitudes of initial position.
    w_theta : 1D array (float)
        Colatitudes of final position.
    w_phi : 1D array (float)
        Longitudes of final position.
    tv_is_vector : bool
        Whether tv contains vectors (i.e. 1st degree tensor).

    Returns
    -------
    1D array of 2D or 1D arrays (float)
        Parallel transported tangent tensors (vectors) at positions w.
    """

    cospsi, sinpsi = get_rotation_angle_array(v_theta, v_phi, w_theta, w_phi)

    if tv_is_vector:
        # compute the parallel transported vectors
        tw = np.zeros_like(tv)
        tw[0] = tv[0] * cospsi + tv[1] * sinpsi
        tw[1] = -tv[0] * sinpsi + tv[1] * cospsi
    else:
        # compute the parallel transported tensors
        tw = rotate_tensor_array(tv, cospsi, sinpsi)
    return tw


############################# AUXYLIARY FUNCTIONS #############################


def get_rotation_angle_array_internal(v_theta, v_phi, w_theta, w_phi):
    N = v_theta.shape[0]

    # convert angles to 3D position vector
    vx = np.sin(v_theta)
    vz = np.cos(v_theta)
    w = hp.ang2vec(w_theta, w_phi - v_phi)
    w = np.ascontiguousarray(w.T)

    # compute axis
    axis = np.empty((3, N))
    axis[0] = -vz * w[1]
    axis[1] = vz * w[0] - vx * w[2]
    axis[2] = vx * w[1]

    cosangle = vx * w[0] + vz * w[2]
    sinangle = np.sqrt(
        axis[0] ** 2 + axis[1] ** 2 + axis[2] ** 2
    )  # np.linalg.norm(axis, axis=0)
    sin0 = np.where(sinangle != 0)
    axis.T[sin0] /= sinangle[sin0][:, np.newaxis]
    axis.T[np.where(sinangle == 0)] = np.array([1.0, 0.0, 0.0])

    # define a vector which lays on the equator and is perpendicular to v
    p = np.empty_like(w)
    p[1] = vx * cosangle
    p[0] = -axis[2] * vx * sinangle
    p[2] = axis[0] * vx * sinangle
    p += axis * (axis[1] * vx * (1 - cosangle))[np.newaxis, :]

    # define ephi_w and etheta_w
    etheta_w0 = w[2] * w[0]
    etheta_w1 = w[2] * w[1]
    etheta_w2 = -(w[0] ** 2 + w[1] ** 2)

    # compute the final rotation angle
    norm = np.sqrt((1.0 - w[2]) * (1.0 + w[2]) * (1.0 - vz) * (1.0 + vz))
    norm0 = np.where(norm == 0.0)
    sinpsi = (p[0] * etheta_w0 + p[1] * etheta_w1 + p[2] * etheta_w2) / norm
    sinpsi[norm0] = 0.0

    cospsi = (p[1] * w[0] - p[0] * w[1]) / norm
    cospsi[norm0] = 1.0
    return cospsi, sinpsi


def get_rotation_angle_array(v_theta, v_phi, w_theta, w_phi):
    """When parallel transporting a tensor of the sphere along a geodesic connecting the
    position v and w, it can be shown that with some strategic choices the whole operation
    can be reduced to a rotation of the tensor by an angle psi. This routine computes
    cosine and sine of such angle.

    Parameters
    ----------
    v_theta : float
        Colatitude of initial position.
    v_phi : float
        Longitude of initial position.
    w_theta : float
        Colatitude of final position.
    w_phi : float
        Longitude of final position.

    Returns
    -------
    1D array (float)
        Cosine of the rotation angles.
    1D array (float)
        Sine of the rotation angles.
    """
    N = v_theta.shape[0]
    bunchsize = 10000
    cospsi = np.empty(N)
    sinpsi = np.empty(N)
    lo = 0
    while lo < N:
        hi = min(N, lo + bunchsize)
        cospsi[lo:hi], sinpsi[lo:hi] = get_rotation_angle_array_internal(
            v_theta[lo:hi], v_phi[lo:hi], w_theta[lo:hi], w_phi[lo:hi]
        )
        lo = hi
    return cospsi, sinpsi


def rotate_tensor_array_internal(t, cospsi, sinpsi):
    msp = -sinpsi
    Rt00 = cospsi * t[0, 0] + sinpsi * t[1, 0]
    Rt01 = cospsi * t[0, 1] + sinpsi * t[1, 1]
    Rt10 = msp * t[0, 0] + cospsi * t[1, 0]
    Rt11 = msp * t[0, 1] + cospsi * t[1, 1]
    t[0, 0] = Rt00 * cospsi + Rt01 * sinpsi
    t[0, 1] = Rt00 * msp + Rt01 * cospsi
    t[1, 0] = Rt10 * cospsi + Rt11 * sinpsi
    t[1, 1] = Rt10 * msp + Rt11 * cospsi


def rotate_tensor_array(t, cospsi, sinpsi):
    """Rotates the tensors contained in t by the angles contained in psi
    via the rotation matrix R.

    R = |  cospsi  sinpsi |
        | -sinpsi  cospsi |

    t_r = R * t * R^-1

    Parameters
    ----------
    t : 1D array of 2D arrays (float)
        Tensors to be rotated.
    cospsi : 1D array (float)
        Cosine of the angles psi.
    sinpsi : 1D array (float)
        Sine of the angles psi.

    Returns
    -------
    1D array of 2D arrays (float)
        Parallel transported tangent tensors (vectors) at positions w.
    """
    N = t.shape[2]
    bunchsize = 5000
    lo = 0
    while lo < N:
        hi = min(N, lo + bunchsize)
        rotate_tensor_array_internal(t[:, :, lo:hi], cospsi[lo:hi], sinpsi[lo:hi])
        lo = hi
    return t



import numpy as np
from scipy import integrate
from scipy.optimize import root_scalar
from .constants import c_cgs


def E(z: float, 
      Omega_M: float, 
      Omega_L: float, 
      Omega_K: float = 0, 
      Omega_R: float = 0):
    """Computes the dimensionless Hubble parameter: E(z) = H/H_0

    Parameters
    ----------
    z : float
        Redshift.
    Omega_M : float
        Matter density, cosmological parameter.
    Omega_L : float
        Cosmological constant density, cosmological parameter.
    Omega_K : float
        Curvature density, cosmological parameter.
    Omega_R : float
        Radiation density, cosmological parameter.

    Returns
    -------
    float
        E(z).
    """
    z_1 = 1 + z
    E_2 = Omega_R * z_1**4 + Omega_M * z_1**3 + Omega_K * z_1**2 + Omega_L
    return np.sqrt(E_2)


def d_c(
    z: float, 
    Omega_M: float, 
    Omega_L: float, 
    Omega_K: float = 0, 
    Omega_R: float = 0
):
    """Computes the comoving distance (in Mpc/h) at the given redshift and the
    cosmological parameters.

    Parameters
    ----------
    z : float
        Redshift.
    Omega_M : float
        Matter density, cosmological parameter.
    Omega_L : float
        Cosmological constant density, cosmological parameter.
    Omega_K : float
        Curvature density, cosmological parameter.
    Omega_R : float
        Radiation density, cosmological parameter.

    Returns
    -------
    float
        Comoving distance (in Mpc/h).
    """
    E_z = lambda x: 1 / E(
        z=x,
        Omega_M=Omega_M,
        Omega_L=Omega_L,
        Omega_K=Omega_K,
        Omega_R=Omega_R,
    )
    dist = integrate.quad(E_z, 0, z)[0]
    dist *= c_cgs / 1e7
    return dist


def z_from_d_c(
    d_c_target: float,
    Omega_M: float,
    Omega_L: float,
    Omega_K: float = 0,
    Omega_R: float = 0,
    z_max: float = 1000,
):
    """Computes redshift given the comoving distance (in Mpc/h) and the
    cosmological parameters. It is the inverse function of d_c.

    Parameters
    ----------
    d_c_target : float
        Redshift.
    Omega_M : float
        Matter density, cosmological parameter.
    Omega_L : float
        Cosmological constant density, cosmological parameter.
    Omega_K : float
        Curvature density, cosmological parameter.
    Omega_R : float
        Radiation density, cosmological parameter.
    z_max : float
        Upper limit to the redshift range, i.e. the value will be
        looked for in the range [0, z_max].

    Returns
    -------
    float
        Redshift.
    """
    f = lambda z: d_c(z, Omega_M, Omega_L, Omega_K, Omega_R) - d_c_target
    return root_scalar(f, bracket=[0.0, z_max]).root



import numpy as np
import healpy as hp


def find_peaks(m, find_minima=False, fout=None):
    """Given a matrix m, returns an array with the values of the peaks found in it.
    A peak is defined as an element which is greater than its 8 neighbors.

    Parameters
    ----------
    m : 2D array
        Image where to look for peaks o minima.
    find_minima : bool
        Wheter to look for the minima instead of the peaks.
    fout : str
        Path of the folder which to save the output.

    Returns
    -------
    1D array
        Array containing the peaks/minima.
    """
    # Get shape of the matrix and prepare list for the peaks
    n, peaks = m.shape[0], []
    # Iterate over rows and columns (except for the borders)
    for i in range(1, n - 1):
        for j in range(1, n - 1):
            # Check that the element is greater/smaller than its 8 neighbors
            flag = 0
            for k in range(-1, 2):
                for l in range(-1, 2):
                    if k == 0 and l == 0:
                        continue
                    if not (find_minima):
                        if m[i, j] <= m[i + k, j + l]:
                            flag = 1
                            break
                    else:
                        if m[i, j] >= m[i + k, j + l]:
                            flag = 1
                            break
                if flag == 1:
                    break
            if flag == 0:
                peaks.append(m[i, j])
    out = np.array(peaks)
    # Eventually save peaks/minima in numpy format
    if fout:
        if not (find_minima):
            fout1 = fout + "_peaks"
        else:
            fout1 = fout + "_minima"
        np.save(fout1, out)
        print(f"Saved extrema to {fout1}")
    return out


def find_peaks_fullsky(m, nest=False, find_minima=False, fout=None):
    """Given a HEALPix map, returns an array with the values of the peaks found
    in it. A peak is defined as an element which is greater than its 8 (or 7)
    HEALPix neighbors.

    Parameters
    ----------
    m : array-like
        HEALPix map where to look for peaks/minima.
    nest : bool
        Whether the angular map is nest ordered or not.
    find_minima : bool
        Wheter to look for the minima instead of the peaks.
    fout : str
        Path of the folder which to save the output.

    Returns
    -------
    1D array
        Array containing the indeces of the peaks/minima.
    1D array
        Array containing the values of the peaks/minima.
    """
    # Get size of the map and prepare list for the peaks and their locations
    npix, vals, locs = len(m), [], []
    nside = hp.npix2nside(npix)
    # Iterate over map
    for ipix in range(npix):
        # Check that the element is greater/smaller than its 8 neighbors
        flag = 0
        nbrs = hp.get_all_neighbours(nside, ipix, nest=nest)
        if -1 in nbrs:
            nbrs = np.delete(nbrs, np.where(nbrs == -1))
        for jpix in nbrs:
            if not (find_minima):
                if m[ipix] <= m[jpix]:
                    flag = 1
                    break
            else:
                if m[ipix] >= m[jpix]:
                    flag = 1
                    break
        if flag == 0:
            vals.append(m[ipix])
            locs.append(ipix)
    # Eventually save peaks/minima in numpy format
    if fout:
        out = np.stack((locs, vals), axis=-1)
        if not (find_minima):
            fout1 = fout + "_peaks"
        else:
            fout1 = fout + "_minima"
        np.save(fout1, out)
        print(f"Saved extrema to {fout1}")
    return locs, vals



import numpy as np
import healpy as hp
from functools import partial


def extract_region(m, lon, lat, phi=0, alpha=5, n=1024, fout=None, nest=False):
    """Extracts a n x n matrix containing the values of the map m inside a square patch.
    The square patch is centered at (lon, lat) and has a side of alpha degrees.

    Parameters
    ----------
    m : array-like
        HEALPix map from which extract the region.
    lon : array-like or int
        Longitude(s) of the maps center(s), in degrees.
    lat : array-like or int
        Latitude(s) of the maps center(s), in degrees.
    phi : array-like or int
        Rotation angle, in degrees.
    alpha : float
        Angular size of the maps, in degrees.
    n : int
        Number of pixels per side of the region.
    fout : str
        Path of the folder which to save the output.
    nest : bool
        Whether the angular maps are nest ordered or not.

    Returns
    -------
    2D array
        The desired region.
    """
    # Compute region
    nside = hp.npix2nside(len(m))
    proj = hp.projector.GnomonicProj(
        rot=(lon, lat, phi), xsize=n, reso=(alpha * 60) / n
    )
    region = proj.projmap(m, vec2pix_func=partial(hp.vec2pix, nside))
    # Eventually save region in numpy format
    if fout:
        print(f"Saved region to {fout}")
        np.save(fout, region)
    return region


def smooth_top_hat(m, radius, nest=False):
    """Smooth an HEALPix map with a top hat filter with circular shape with
    given radius.

    Parameters
    ----------
    m : array-like
        HEALPix map to smooth.
    radius : float
        Radius of the filter (in radians).
    nest : bool
        Whether the angular maps are nest ordered or not.

    Returns
    -------
    array-like
        The smoothed map.
    """
    npix = m.shape[0]
    nside = hp.npix2nside(npix)
    m_out = np.zeros(npix, dtype=m.dtype)
    for ipix in range(npix):
        theta, phi = hp.pix2ang(nside, ipix, nest=nest)
        vec = hp.ang2vec(theta, phi)
        disc = hp.query_disc(nside, vec, radius, nest=nest)
        m_out[ipix] = np.mean(m[disc])
    return m_out


def rotate_map(m, rot_theta, rot_phi, nest=False):
    """Take a healpix map m and returns another map healpix rotated
    by the (theta, phi) given.

    Parameters
    ----------
    m : array-like
        HEALPix map to rotate.
    rot_theta : float
        Rotation angle theta in radians.
    rot_phi : float
        Rotation angle phi in radians.
        nest : bool
        Whether the angular maps are nest ordered or not.

    Returns
    -------
    array-like
        Rotated map.
    """
    nside = hp.npix2nside(len(m))
    # Get theta, phi for non-rotated map
    t, p = hp.pix2ang(nside, np.arange(hp.nside2npix(nside)))  # theta, phi
    # Define a rotator
    r = hp.Rotator(deg=False, rot=[rot_phi, rot_theta])
    # Get theta, phi under rotated co-ordinates
    trot, prot = r(t, p)
    # Interpolate map onto these co-ordinates
    rot_map = hp.get_interp_val(m, trot, prot, nest=nest)
    return rot_map


def ud_grade_interp(map_in, nside_out, nest=False, interpolation="bilinear"):
    """Up or downgrades an healpix map to a map nside = nside_out.
    The values in the output map are computed via either NGP or bilinear
    interpolation.

    Parameters
    ----------
    map_in : array-like
        HEALPix map to up/down-grade.
    nside_out : int
        NSIDE of the output map.
    nest : bool
        Whether the angular maps are nest ordered or not.
    interp : str
        Interpolation scheme to use: "ngp" or "bilinear".


    Returns
    -------
    array-like
        Up/downgraded map.
    """
    nside_in = hp.npix2nside(map_in.shape[0])
    npix_out, npix_in = hp.nside2npix(nside_out), hp.nside2npix(nside_in)
    map_out = np.zeros(npix_out, dtype=map_in.dtype)
    theta, phi = hp.pix2ang(nside_out, np.array(range(npix_out)), nest=nest)
    if interpolation == "bilinear":
        map_out = hp.get_interp_val(map_in, theta, phi, nest=nest)
    elif interpolation == "ngp":
        map_out = map_in[hp.ang2pix(nside_in, theta, phi, nest=nest)]
    else:
        print("interpolation method not valid")
        return
    map_out /= npix_out / npix_in
    return map_out


def get_rings_idx(nside, get_colatitude=False):
    """Computes the starting index of each HEALPix ring, given a certain Nside.

    Parameters
    ----------
    nside : int
        HEALPix Nside parameter.
    get_colatitude : bool
        Wheter to also get the co-latitude of the respective rings.

    Returns
    -------
    array-like
        Indices of the rings.
    """
    rings_idx = []
    npix = hp.nside2npix(nside)
    ncap = 2 * nside * (nside - 1)
    # north polar cap
    for i_z in range(1, nside):
        i_ring = i_z
        ipix = 2 * i_ring * (i_ring - 1)
        rings_idx.append(ipix)
    # equatorial region
    for i_z in range(nside, 3 * nside + 1):
        i_ring = i_z - nside + 1
        ipix = ncap + 4 * nside * (i_ring - 1)
        rings_idx.append(ipix)
    # south polar cap
    for i_z in range(3 * nside + 1, 4 * nside):
        i_ring = 4 * nside - i_z
        ipix = npix - 2 * i_ring * (i_ring + 1)
        rings_idx.append(ipix)
    rings_idx = np.array(sorted(rings_idx))
    if not (get_colatitude):
        return rings_idx
    else:
        colatitude = hp.pix2ang(nside, rings_idx)
        return rings_idx, colatitude[0]


def shot_noise(nshot, nside):
    """Fills the healpix map "m" (with the chosen nside) with a number of shots
    drawn from a uniform distribution, so to obtain Poisson noise.

    Parameters
    ----------
    nshot : int
        Number of shots to draw.
    nside : int
        HEALPix Nside parameter.

    Returns
    -------
    array-like
        HEALPix noise map'.
    """
    npix = 12 * nside**2
    m = np.zeros(npix)
    for i in range(nshot):
        m[np.random.randint(0, npix)] += 1
    return m


def IndexToDecRa(index, nside):
    """Converts HEALPix index to corresponding Re and Dec.

    Parameters
    ----------
    index : int
        HEALPix index.
    nside : int
        HEALPix Nside.

    Returns
    -------
    Float
        Declination.
    Float
        RIght ascension.
    """
    theta, phi = hp.pixelfunc.pix2ang(nside, index)
    return -np.degrees(theta - np.pi / 2.0), np.degrees(np.pi * 2.0 - phi)


def alm2map_der2(alm, nside, lmax=None, mmax=None):
    """Computes a Healpix map and its second derivatives given the alm.

    The alm are given as a complex array. You can specify lmax
    and mmax, or they will be computed from array size (assuming
    lmax==mmax).
    
    The notations of kappa and gamma are borrowed from the weak lensing
    formalism. See e.g. eq (7, 8, 9, 25, 26) of 
    (Castro et al. 2005, Phys. Rev. D, 72, 023516)

    Parameters
    ----------
    alm : array, complex
      A complex array of alm. Size must be of the form mmax(lmax-mmax+1)/2+lmax
    nside : int
      The nside of the output map.
    lmax : None or int, optional
      Explicitly define lmax (needed if mmax!=lmax)
    mmax : None or int, optional
      Explicitly define mmax (needed if mmax!=lmax)

    Returns
    -------
    m, d_theta^2, d_theta*d_phi , d_phi^2 : tuple of arrays
      The maps correponding to alm, and its second derivatives with respect 
      to theta and phi. d_theta*d_phi is already divided by sin(theta) and
      d_phi^2 is already divided by sin^2(theta)
    """
    if lmax is None: 
        lmax = 3 * nside
        mmax = lmax
    m = hp.alm2map(alm, nside=nside, lmax=lmax, mmax=mmax)
    ell = np.arange(0, lmax + 1)    
    f_l = np.sqrt((ell + 2.0) * (ell - 1.0) / (ell * (ell + 1.0)))
    kappa_lm = hp.almxfl(alm, -0.5*(ell * (ell + 1)))
    g_lm = hp.almxfl(kappa_lm, f_l)
    kappa = hp.alm2map(kappa_lm, nside=nside, lmax=lmax, mmax=mmax)
    g1, g2 = hp.alm2map_spin(
        [g_lm, np.zeros_like(g_lm)], nside=nside, spin=2, lmax=lmax, mmax=mmax
    )
    return m, kappa + g1, g2, kappa - g1


import numpy as np
import healpy as hp
from .cosmology import d_c
from .raytracing import raytrace
from .logging import info, success, warning
from math import prod


def _single_raytrace(shells, z_s, omega_m, omega_l, nside,
                     redshifts_clean, distances,
                     interp, lmax, parallel_transport, nufft_nthreads):
    """Run ray-tracing for a single source redshift and return a result dict."""
    kappa_born, A_final, beta_final, theta = raytrace(
        shells=shells,
        z_s=z_s,
        omega_m=omega_m,
        omega_l=omega_l,
        nside=nside,
        shell_redshifts=redshifts_clean,
        shell_distances=distances,
        interp=interp,
        lmax=lmax,
        parallel_transport=parallel_transport,
        nthreads=nufft_nthreads,
    )
    kappa_raytraced = 1.0 - 0.5 * (A_final[0, 0] + A_final[1, 1])
    n_used = sum(1 for z in redshifts_clean if z < z_s)
    return {
        'convergence_born': kappa_born,
        'convergence_raytraced': kappa_raytraced,
        'distortion_matrix': A_final,
        'ray_positions': beta_final,
        'initial_positions': theta,
        'shell_info': {
            'redshifts': redshifts_clean,
            'distances': distances,
            'n_shells_total': len(redshifts_clean),
            'n_shells_used': n_used,
        },
    }


def _stack_results(results):
    """Stack a list of single-source result dicts into arrays with a leading n_sources axis."""
    return {
        'convergence_born':      np.stack([r['convergence_born']      for r in results]),
        'convergence_raytraced': np.stack([r['convergence_raytraced'] for r in results]),
        'distortion_matrix':     np.stack([r['distortion_matrix']     for r in results]),
        'ray_positions':         np.stack([r['ray_positions']         for r in results]),
        'initial_positions':     np.stack([r['initial_positions']     for r in results]),
        'shell_info':            [r['shell_info'] for r in results],
    }


def prepare_density_shells(
    density_maps,
    redshifts,
    box_size,
    n_particles,
    omega_m,
    h=0.6766,
    omega_l=None,
    nside=None,
    shell_widths=None,
):
    """
    Convert density planes to Dorian mass format for ray-tracing.

    This function handles the conversion from various density representations
    (particle number density, overdensity, or mass) to the mass per pixel format
    expected by Dorian's ray-tracing algorithm.

    Parameters
    ----------
    density_maps : list of array-like
        List of HEALPix density maps for each shell. Each map should be a 1D
        array with ``npix = 12 * nside**2`` elements in RING ordering.
    redshifts : array-like
        Redshift of each shell, must have same length as ``density_maps``.
    box_size : float or tuple of float
        Simulation box size in Mpc/h. If a single value, assumes cubic box.
        If tuple, should be ``(Lx, Ly, Lz)``.
    n_particles : int
        Total number of particles in the simulation (e.g., ``512**3``).
    omega_m : float
        Matter density parameter (Omega_m).
    h : float, optional
        Dimensionless Hubble parameter H0/(100 km/s/Mpc). Default: 0.6766.
    omega_l : float, optional
        Dark energy density parameter (Omega_Lambda). Default: ``1 - omega_m``.
    nside : int, optional
        Target HEALPix NSIDE resolution. If None, auto-detected from first map.
    shell_widths : array-like, optional
        Shell thickness d_R per shell in Mpc/h. Required when density_maps
        contain number density (particles per (Mpc/h)^3). When provided, the
        pixel volume is computed per shell to convert density to mass. When
        None, density_maps are assumed to already contain particle counts per
        pixel and are directly multiplied by the particle mass.

    Returns
    -------
    shells : list of np.ndarray
        Mass per pixel in units of 10^10 M_sun/h, ready for ray-tracing.
    shell_distances : list of float
        Comoving distances to each shell in Mpc/h.
    shell_redshifts : list of float
        Validated redshifts for each shell.

    Examples
    --------
    Basic usage with number density maps and shell widths:

    >>> import numpy as np
    >>> from dorian.lensing import prepare_density_shells
    >>> # Create mock density maps (4 shells, nside=64)
    >>> nside = 64
    >>> npix = 12 * nside**2
    >>> density_maps = [np.random.randn(npix) * 0.01 for _ in range(4)]
    >>> redshifts = [0.1, 0.2, 0.3, 0.4]
    >>> shell_widths = [100.0, 100.0, 100.0, 100.0]  # Mpc/h
    >>> shells, distances, z_out = prepare_density_shells(
    ...     density_maps=density_maps,
    ...     redshifts=redshifts,
    ...     box_size=1000.0,        # Mpc/h
    ...     n_particles=256**3,
    ...     omega_m=0.3,
    ...     shell_widths=shell_widths,
    ... )
    >>> len(shells)
    4

    See Also
    --------
    raytrace_from_density : Combines preparation and ray-tracing in one call.
    dorian.raytracing.raytrace : Low-level ray-tracing function.
    """
    if omega_l is None:
        omega_l = 1.0 - omega_m

    assert shell_widths is not None, "shell_widths must be provided when density_maps contain number density"

    if isinstance(box_size, (int, float)):
        box_size = (box_size, box_size, box_size)

    if nside is None:
        nside = hp.npix2nside(len(density_maps[0]))

    npix = hp.nside2npix(nside)

    rho_crit_h2 = 2.775e11
    rho_crit = rho_crit_h2 * h**2
    rho_matter = omega_m * rho_crit
    volume_box_mpc = prod(box_size) / h**3
    particle_mass_msun = (rho_matter * volume_box_mpc) / n_particles
    particle_mass_dorian = (particle_mass_msun * h) / 1e10

    info(f"Preparing {len(density_maps)} density shells: nside={nside}, "
         f"particle mass={particle_mass_dorian:.4e} [10^10 M_sun/h]")

    shells = []
    shell_distances = []
    shell_redshifts = []

    for i, (density_map, z) in enumerate(zip(density_maps, redshifts)):
        map_nside = hp.npix2nside(len(density_map))
        if map_nside != nside:
            warning(f"Shell {i}: ud_grading nside {map_nside} -> {nside}")
            density_map = hp.ud_grade(density_map, nside, power=-2)

        d_k = d_c(z=z, Omega_M=omega_m, Omega_L=omega_l)
        info(f"  Shell {i+1}/{len(density_maps)}: z={z:.4f}, d={d_k:.1f} Mpc/h")

        dr = shell_widths[i]
        R_min = max(d_k - dr / 2, 0.0)
        R_max = d_k + dr / 2
        shell_vol = (4.0 / 3.0) * np.pi * (R_max**3 - R_min**3)
        pixel_vol = shell_vol / npix
        mass_per_pixel = density_map * pixel_vol * particle_mass_dorian

        shells.append(np.asarray(mass_per_pixel, dtype=np.float64))
        shell_distances.append(float(d_k))
        shell_redshifts.append(float(z))

    success(f"Prepared {len(shells)} density shells")

    return shells, shell_distances, shell_redshifts


def raytrace_from_density(
    density_maps,
    redshifts,
    z_source,
    box_size,
    n_particles,
    omega_m,
    h=0.6766,
    omega_l=None,
    nside=None,
    interp='bilinear',
    shell_widths=None,
    parallel_transport=True,
    lmax=0,
    nufft_nthreads=1,
    comm=None,
):
    """
    Perform full-sky weak lensing ray-tracing from density maps.

    This is the main high-level interface for Dorian. It combines density shell
    preparation and ray-tracing into a single call, making it easy to compute
    weak lensing convergence maps from simulation lightcones.

    Parameters
    ----------
    density_maps : list of array-like
        List of HEALPix density maps for each shell. Each map should be a 1D
        array with ``npix = 12 * nside**2`` elements in RING ordering.
    redshifts : array-like
        Redshift of each shell, must have same length as ``density_maps``.
        Only shells with ``z < z_source`` will be used.
    z_source : float or list of float
        Source redshift(s). When a scalar, traces rays for a single source and
        returns a single dict. When a list, runs each source sequentially (or
        distributed across MPI ranks when ``comm`` is provided), returning a
        merged dict with stacked arrays of shape ``(n_sources, ...)``.
    box_size : float or tuple of float
        Simulation box size in Mpc/h. If a single value, assumes cubic box.
    n_particles : int
        Total number of particles in the simulation (e.g., ``512**3``).
    omega_m : float
        Matter density parameter (Omega_m).
    h : float, optional
        Dimensionless Hubble parameter H0/(100 km/s/Mpc). Default: 0.6766.
    omega_l : float, optional
        Dark energy density parameter (Omega_Lambda). Default: ``1 - omega_m``.
    nside : int, optional
        Target HEALPix NSIDE resolution. If None, auto-detected from first map.
    interp : {'bilinear', 'ngp', 'nufft'}, optional
        Interpolation method for sampling deflection fields at ray positions:

        - ``'bilinear'``: Bilinear interpolation (default, good balance)
        - ``'ngp'``: Nearest grid point (fastest, lowest accuracy)
        - ``'nufft'``: Non-uniform FFT (highest accuracy, slowest)

    shell_widths : array-like, optional
        Shell thickness d_R per shell in Mpc/h. Required when density_maps
        contain number density (particles per (Mpc/h)^3). Passed through to
        ``prepare_density_shells``.
    parallel_transport : bool, optional
        Whether to apply parallel transport of the distortion matrix along
        geodesics. Recommended to keep True for accurate results. Default: True.
    lmax : int, optional
        Maximum multipole ell for spherical harmonic transforms.
        Default: ``3 * nside`` (sufficient for most applications).
    nufft_nthreads : int, optional
        Number of OpenMP threads for ``'nufft'`` interpolation. Default: 1.
    comm : mpi4py.MPI.Comm or None, optional
        An MPI communicator (e.g. ``MPI.COMM_WORLD``). When provided, sources
        are distributed across ranks via round-robin assignment: rank *i*
        handles source indices *i*, *i+size*, *i+2×size*, …. Non-root ranks
        return ``None``; rank 0 returns the full stacked result. When ``None``
        (default), all sources are processed sequentially in the calling
        process.

    Returns
    -------
    results : dict or None
        Dictionary containing all ray-tracing outputs. Non-root MPI ranks
        return ``None``.

        **Scalar z_source** (single-source):

        - ``'convergence_born'`` : np.ndarray, shape ``(npix,)``
        - ``'convergence_raytraced'`` : np.ndarray, shape ``(npix,)``
        - ``'distortion_matrix'`` : np.ndarray, shape ``(2, 2, npix)``
        - ``'ray_positions'`` : np.ndarray, shape ``(2, npix)``
        - ``'initial_positions'`` : np.ndarray, shape ``(2, npix)``
        - ``'shell_info'`` : dict with keys ``'redshifts'``, ``'distances'``,
          ``'n_shells_total'``, ``'n_shells_used'``.

        **List z_source** (multi-source path):

        - ``'convergence_born'`` : np.ndarray, shape ``(n_sources, npix)``
        - ``'convergence_raytraced'`` : np.ndarray, shape ``(n_sources, npix)``
        - ``'distortion_matrix'`` : np.ndarray, shape ``(n_sources, 2, 2, npix)``
        - ``'ray_positions'`` : np.ndarray, shape ``(n_sources, 2, npix)``
        - ``'initial_positions'`` : np.ndarray, shape ``(n_sources, 2, npix)``
        - ``'shell_info'`` : list of dict, one per source.

    Examples
    --------
    Basic ray-tracing from a simulation lightcone:

    >>> import numpy as np
    >>> from dorian.lensing import raytrace_from_density
    >>> # Load your density maps (list of HEALPix maps)
    >>> density_maps = [...]  # 4 shells at different redshifts
    >>> redshifts = [0.1, 0.2, 0.3, 0.4]
    >>> results = raytrace_from_density(
    ...     density_maps=density_maps,
    ...     redshifts=redshifts,
    ...     z_source=1.0,
    ...     box_size=2000.0,       # Mpc/h
    ...     n_particles=512**3,
    ...     omega_m=0.3,
    ...     nside=512,
    ... )
    >>> kappa_born = results['convergence_born']
    >>> kappa_raytraced = results['convergence_raytraced']

    Extract shear from the distortion matrix:

    >>> A = results['distortion_matrix']
    >>> gamma1 = 0.5 * (A[0, 0] - A[1, 1])  # shear component 1
    >>> gamma2 = A[0, 1]                     # shear component 2

    Using high-precision NUFFT interpolation:

    >>> results = raytrace_from_density(
    ...     density_maps=density_maps,
    ...     redshifts=redshifts,
    ...     z_source=1.0,
    ...     box_size=2000.0,
    ...     n_particles=512**3,
    ...     omega_m=0.3,
    ...     interp='nufft',
    ...     nufft_nthreads=8,
    ... )

    Multi-source MPI ray-tracing (outputs have an extra leading axis):

    >>> from mpi4py import MPI
    >>> z_sources = [0.5, 1.0, 2.0]
    >>> results = raytrace_from_density(
    ...     density_maps=density_maps,
    ...     redshifts=redshifts,
    ...     z_source=z_sources,
    ...     box_size=2000.0,
    ...     n_particles=512**3,
    ...     omega_m=0.3,
    ...     nside=512,
    ...     comm=MPI.COMM_WORLD,   # distribute across MPI ranks
    ... )
    >>> # On rank 0:
    >>> results['convergence_born'].shape   # (3, npix)
    (3, 3145728)
    >>> results['distortion_matrix'].shape  # (3, 2, 2, npix)
    (3, 2, 2, 3145728)

    Notes
    -----
    The convergence is computed two ways:

    1. **Born approximation** (``convergence_born``): Integrates kappa along
       unperturbed (straight) ray paths. Fast but ignores lens-lens coupling.

    2. **Ray-traced** (``convergence_raytraced``): Computed from the full
       distortion matrix as ``kappa = 1 - 0.5 * Tr(A)``. Includes all
       post-Born corrections.

    See Also
    --------
    prepare_density_shells : Convert density maps to mass format separately.
    dorian.raytracing.raytrace : Low-level ray-tracing function.
    """
    if omega_l is None:
        omega_l = 1.0 - omega_m

    # Detect multi-source mode using robust scalar and sequence checks.
    # - Scalars: np.isscalar(...) or 0-D numpy arrays -> single source.
    # - Multi-source: explicit sequences (list/tuple) or numpy arrays with ndim > 0.
    if np.isscalar(z_source) or (isinstance(z_source, np.ndarray) and z_source.ndim == 0):
        is_multi = False
    elif isinstance(z_source, (list, tuple, np.ndarray)):
        # For numpy arrays, ndim > 0 is guaranteed here since 0-D case is handled above.
        is_multi = True
    else:
        # Fallback: treat unknown types as single source to avoid misclassification.
        is_multi = False

    info(f"raytrace_from_density: z_source={z_source}, interp='{interp}', "
         f"nside={nside if nside else 'auto'}, multi={is_multi}")

    # ── Prepare Density Shells (rank 0 only, then broadcast) ───────────────
    if comm is not None:
        rank = comm.Get_rank()
        if rank == 0:
            prepared_data = prepare_density_shells(
                density_maps=density_maps,
                redshifts=redshifts,
                box_size=box_size,
                n_particles=n_particles,
                omega_m=omega_m,
                h=h,
                omega_l=omega_l,
                nside=nside,
                shell_widths=shell_widths,
            )
        else:
            prepared_data = None
        shells, distances, redshifts_clean = comm.bcast(prepared_data, root=0)
    else:
        shells, distances, redshifts_clean = prepare_density_shells(
            density_maps=density_maps,
            redshifts=redshifts,
            box_size=box_size,
            n_particles=n_particles,
            omega_m=omega_m,
            h=h,
            omega_l=omega_l,
            nside=nside,
            shell_widths=shell_widths,
        )

    if nside is None:
        nside = hp.npix2nside(len(shells[0]))

    # Normalize z_source to list internally
    z_sources = list(z_source) if is_multi else [float(z_source)]

    if comm is not None:
        # ── MPI path ──────────────────────────────────────────────────────────
        rank = comm.Get_rank()
        size = comm.Get_size()

        import dorian.logging as _dlog
        _dlog._prefix = f"[rank {rank}] "

        # Round-robin: rank i handles indices i, i+size, i+2*size, ...
        my_indices = list(range(rank, len(z_sources), size))
        my_results = []
        for idx in my_indices:
            z = z_sources[idx]
            info(f"Processing z_source={z:.4f}")
            result = _single_raytrace(shells, z, omega_m, omega_l, nside,
                                      redshifts_clean, distances,
                                      interp, lmax, parallel_transport, nufft_nthreads)
            my_results.append((idx, result))

        # Gather all (index, result) pairs to rank 0
        gathered = comm.gather(my_results, root=0)
        # barrier
        comm.Barrier()

        if rank == 0:
            # Flatten and sort by original z_source order
            flat = sorted(
                [pair for rank_list in gathered for pair in rank_list],
                key=lambda x: x[0],
            )
            ordered_results = [r for _, r in flat]

            success(f"raytrace_from_density complete: {len(z_sources)} sources across {size} ranks")

            if not is_multi:
                return ordered_results[0]
            return _stack_results(ordered_results)
        else:
            return None  # non-root ranks return None

    else:
        # ── Sequential path (no MPI) ──────────────────────────────────────────
        ordered_results = []
        for z in z_sources:
            ordered_results.append(
                _single_raytrace(shells, z, omega_m, omega_l, nside,
                                 redshifts_clean, distances,
                                 interp, lmax, parallel_transport, nufft_nthreads)
            )
        success(f"raytrace_from_density complete: {len(z_sources)} sources (sequential)")

        if not is_multi:
            return ordered_results[0]
        return _stack_results(ordered_results)



import os, h5py
import numpy as np
from pathlib import Path
import healpy as hp
from .healpix import ud_grade_interp


def load_massmap(simDir, nmap):
    """Load the mass from the mass shell output of Gadget4.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder,
        e.g. '/virgo/simulations/MTNG/DM-Gadget4/MTNG-L500-270-A'.
    nmap : int
        Number of the mass-shell to load.

    Returns
    -------
    array-like
        HEALPix map containing the mass data.
    """
    mDir = f"{simDir}/mapsdir_{nmap:03d}/"
    offset = 0
    header = load_header(simDir, nmap)
    tot_pix, num_files = header["NpixTotal"], header["NumFiles"]
    mass = np.zeros(tot_pix, dtype=np.float64)
    for i in range(num_files):
        with h5py.File(f"{mDir}maps_{nmap:03d}.{i}.hdf5", "r") as f:
            m_temp = f["Maps"]["Mass"]
            # npix = header["NpixLocal"]
            # this should be the correct line, but sometimes NpixLocal is wrong
            npix = m_temp.shape[0]
            mass[offset : offset + npix] = m_temp
            offset += npix
    return mass


def load_header(simDir, nmap):
    """Load the header from the mass shell output of Gadget4.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder,
        e.g. '/virgo/simulations/MTNG/DM-Gadget4/MTNG-L500-270-A'.
    nmap : int
        Number of the mass-shell to load.

    Returns
    -------
    dict
        Dictionary containing the header. Keys are: Nmap, Redshift, AscaleStart,
        ComDistEnd, ComDistStart, NpixLocal, NpixTotal, Nside, NumFiles
    """
    with h5py.File(f"{simDir}/mapsdir_{nmap:03d}/maps_{nmap:03d}.0.hdf5", "r") as f:
        h = f["Header"]
        out = {}
        for k in h.attrs.keys():
            out[k] = h.attrs[k]
        out["Nmap"] = nmap
        out["Redshift"] = 1 / ((out["AscaleEnd"] + out["AscaleStart"]) * 0.5) - 1
    return out


def load_all_headers(simDir):
    """Load all the mass-shell info."""
    out = []
    for file in os.listdir(simDir):
        fname = os.fsdecode(file)
        if fname.startswith("mapsdir"):
            n = int(fname[-3:])
            out.append(load_header(simDir, nmap=n))
    out.sort(key=lambda x: x["Nmap"])
    return out


def load_params(simDir):
    """Load the parameters from the mass shell output of Gadget4.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder,
        e.g. '/virgo/simulations/MTNG/DM-Gadget4/MTNG-L500-270-A'.

    Returns
    -------
    dict
        Dictionary containing the parameters.
    """
    with h5py.File(f"{simDir}/mapsdir_000/maps_000.0.hdf5", "r") as f:
        h = f["Parameters"]
        params = {}
        for k in h.attrs.keys():
            params[k] = h.attrs[k]
    return params


def merge_massmaps(simDir, outDir, n_merge):
    """Merges every n mass-shells present in simDir, starting from the innermost one.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder.
    outDir : str
        Path where to save the new mass-shells.
    n_merge : int
        Number of the mass-shells to merge togehter.
    """
    print(f"Merging together every {n_merge} mass-shells form {simDir} to {outDir}")
    Path(f"{outDir}").mkdir(exist_ok=True)
    sh_tot = 0
    for f in os.listdir(simDir):
        if os.fsdecode(f).startswith("mapsdir"):
            sh_tot += 1
    h = load_header(simDir, 0)
    npix = h["NpixTotal"]
    mass_to_save, n_shells_cumulative = np.zeros(npix), 0
    ComDistEnd, AscaleEnd = 0, 1.0
    print(f"sh_tot: {sh_tot}")
    for n_shell in range(sh_tot - 1, -1, -1):
        n_shells_cumulative += 1
        h = load_header(simDir, n_shell)
        ComDistStart, AscaleStart = h["ComDistStart"], h["AscaleStart"]
        print(f"ComDistStart, AscaleStart: {ComDistStart, AscaleStart}")
        mass = load_massmap(simDir, n_shell)
        mass_to_save += mass
        if n_shells_cumulative % n_merge == 0:
            h = load_header(simDir, n_shell)
            ComDistStart, AscaleStart = h["ComDistStart"], h["AscaleStart"]
            print(f"Merging the above {n_merge} shells...")
            Path(f"{outDir}/mapsdir_{n_shell//n_merge:03d}").mkdir(exist_ok=True)
            with h5py.File(
                f"{simDir}/mapsdir_{n_shell:03d}/maps_{n_shell:03d}.0.hdf5", "r"
            ) as f_in:
                n_fac = n_shell // n_merge
                with h5py.File(
                    f"{outDir}/mapsdir_{n_fac:03d}/maps_{n_fac:03d}.0.hdf5", "w"
                ) as f_out:
                    for k1 in f_in.keys():
                        f_in.copy(k1, f_out)
                    del f_out["Maps/Mass"]
                    f_out["Maps"].create_dataset("Mass", data=mass_to_save, dtype="f4")
                    f_out["Header"].attrs["ComDistStart"] = ComDistStart
                    f_out["Header"].attrs["ComDistEnd"] = ComDistEnd
                    f_out["Header"].attrs["AscaleStart"] = AscaleStart
                    f_out["Header"].attrs["AscaleEnd"] = AscaleEnd
                    f_out["Header"].attrs["NpixLocal"] = f_out["Header"].attrs[
                        "NpixTotal"
                    ]
                    f_out["Header"].attrs["NumFiles"] = 1
            mass_to_save, n_shells_cumulative = np.zeros(npix), 0
            ComDistEnd, AscaleEnd = ComDistStart, AscaleStart
    print("Finished to merge the mass-shells, bye!")


def downsample_massmap(
    simDir,
    outDir,
    nside_out,
    n_start,
    n_end,
    nest=False,
    interpolation="bilinear",
    use_healpix_routine=False,
):
    """Downsamples the mass-shells present in simDir from n_start to n_end.

    Parameters
    ----------
    simDir : str
        Path to the simulation output folder.
    outDir : str
        Path where to save the new mass-shells.
    nside_out : int
        Healpix Nside of the downsampled mass-shells.
    interp : str
        Interpolation scheme to use, used only if use_healpix_routine is False
    use_healpix_routine : boolean
        If True, healpy ud_grade routine will be used
    """
    print(
        f"Downsampling to nside {nside_out} mass-shells from n.{n_start} to n.{n_end}.\n"
        + f"Simulation folder: {simDir}\n"
        + f"Output folder: {outDir}\n",
        flush=True,
    )

    Path(f"{outDir}").mkdir(exist_ok=True)
    for nmap in range(n_start, n_end + 1):
        mass = load_massmap(simDir, nmap)
        if use_healpix_routine:
            order = "NEST" if nest else "RING"
            mass_downsampled = hp.ud_grade(
                mass, nside_out, order_in=order, order_out=order, power=-2
            )
        else:
            mass_downsampled = ud_grade_interp(mass, nside_out, nest, interpolation)

        Path(f"{outDir}/mapsdir_{nmap:03d}").mkdir(exist_ok=True)

        with h5py.File(
            f"{simDir}/mapsdir_{nmap:03d}/maps_{nmap:03d}.0.hdf5", "r"
        ) as f_in:
            with h5py.File(
                f"{outDir}/mapsdir_{nmap:03d}/maps_{nmap:03d}.0.hdf5", "w"
            ) as f_out:
                for k1 in f_in.keys():
                    f_in.copy(k1, f_out)
                del f_out["Maps/Mass"]
                f_out["Maps"].create_dataset("Mass", data=mass_downsampled, dtype="f4")
                h_out = f_out["Header"]
                h_out.attrs["NpixLocal"] = hp.nside2npix(nside_out)
                h_out.attrs["NpixTotal"] = hp.nside2npix(nside_out)
                h_out.attrs["Nside"] = nside_out
                h_out.attrs["NumFiles"] = 1
        print(f"Done with massmap n. {nmap}", flush=True)

    print("Finished to downsample the mass-shells, bye!")



def write_massmap(
    outDir,
    nmap,
    mass_array,
    ComDistStart,
    ComDistEnd,
    Omega_M,
    Omega_L,
    HubbleParam,
):
    """Converts a HALPix map (given e.g. as a numpy array) in the Gadget4 massmap format,
    adopted by Dorian.

    Parameters
    ----------
    outDir : str
        Path where to save the massmaps.
    nmap : int
        Number of the massmap. These should be 0 for the outtermost shell and increasing
        as it reaches the observer (innermost shell)
    mass_array : array-like
        HEALPix map containing the binned mass, in units of 1e10*M_sun/h
    ComDistStart; float
        Outter boundary of the massmap, in Mpc/h
    ComDistEnd; float
        Inner boundary of the massmap, in Mpc/h
    Omega_M : float
        Matter density, cosmological parameter.
    Omega_L : float
        Cosmological constant density, cosmological parameter.
    HubbleParam : float
        Reduced Hubble constant, cosmological parameter.
    """
    Npix = len(mass_array)
    Path(f"{outDir}/mapsdir_{nmap:03d}").mkdir(exist_ok=True)
    with h5py.File(
        f"{outDir}/mapsdir_{nmap:03d}/maps_{nmap:03d}.0.hdf5", "w"
    ) as f_out:
        h_out = f_out.create_group("Header")
        h_out.attrs["ComDistStart"] = ComDistStart
        h_out.attrs["ComDistEnd"] = ComDistEnd
        h_out.attrs["NpixLocal"] = Npix
        h_out.attrs["NpixTotal"] = Npix
        h_out.attrs["Nside"] = hp.npix2nside(Npix)
        h_out.attrs["NumFiles"] = 1
        
        param_out = f_out.create_group("Parameters")
        param_out.attrs["Omega0"] = Omega_M
        param_out.attrs["OmegaLambda"] = Omega_L
        param_out.attrs["HubbleParam"] = HubbleParam
        
        f_out.create_group("Maps")
        f_out["Maps"].create_dataset("Mass", data=mass_array, dtype="f4")